From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq.
From Kont Require Import rels domains prelude heaps hprop.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

(************************************)
(* HTTcc Algebraic call/cc model    *)
(************************************)

(* The Kont type makes explicit the connection with the jumping point:  *)

(* Definition Kont A  (s : spec A) : Type :=                                *)
(*   forall k,                                                                *)
(*     SK _ (fun i => s.1 i /\ i = k, s.2) -> def k -> s.1 k ->  X.  *)

(* callccR : the one tha we've been using for the last couple of months,      *)
(*  where R describes the relationship between the initial(captured) heap and *)
(* the heap where the jumps occurs, i.e. the one that is passed to the        *)
(* continuation object when an abort is evaluated.                            *)

(*  TT j.  f : Stcont A (fun i => R j i /\ pf i, qf) |- e : SK a (pe j, qe j) *)
(* -------------------------------------------------------------------------- *)
(*  calccR R : SK A (fun i=> pe i i /\ TT s. R i s -> pf s, *)
(*                  (fun y i m=> qe i y i m \/ (R ; qf) y i m) *)

(*         g : SKCont A (pg,qg)        t : SK A (pt,qt) *)
(* -------------------------------------------------------------------------- *)
(* abort  g t : _/\_ (B : Type).                                              *)
(*    SK B (fun i=> pg i /\ (pg,qg) <= (fun j =>pt i /\ j = i,qt) , False)    *)


Delimit Scope ccscope with cc. 
Open Scope cc.
Definition pre := heap -> Prop.
Definition post A := A -> heap -> heap -> Prop.
Definition spec A := (pre * post A)%type.

Definition prog (R: lattice) A  (s : spec A) := 
  forall i, def i -> s.1 i -> 
    (forall y m, def m -> s.2 y i m ->  R) ->  R.

Structure SK R A (s : spec A) : Type := Prog {
  model : prog R s}.

Arguments model {R A  s} _ _ _ _ _ .

Arguments SK R {A} s .

(* The type of continuations is different from traditional callcces *)
(* This is due to a different signature for callcc: *)

(* Sigma_CC A = ( A -> S \bot ) -> A *)

(* Translated to the model this means that the SKCont has type: *)
(* SK s -> heap -> X. *)

(* However, we can take advantage of the DT setting and make explicit that *)
(* the spec depends on the heap *)

Definition Kont X A  (s : spec A) : Type := 
    forall k, 
      SK X  (fun i => s.1 i /\ i = k, s.2) -> def k -> s.1 k ->  X.

Arguments Kont X {A} s.


(* Sigma Types for nesting continuations *)


(* In order  to be able to flatten pre and post conditions when returning     *)
(* computations will attempt to define it as a particular return type which   *)
(* carries its pre and post conditions, by means of a dependent sum type:     *)

Inductive SigSK R (A : Type) := BigSK {spec_of : spec A; comm: @SK R A spec_of}.

Arguments SigSK R {A}. 


(**************************)
(* Poset Structure for SK *)
(**************************)

(* Two computations are equal if they have the same model:    *)

Lemma modelE R A (s: spec A) (e1 e2: SK R  s) :
        e1 = e2 <-> model e1 = model e2.
Proof. by move: e1 e2=>[e1][e2]; split=>/= [[]|]->//. Qed.

Section SKPos.
Variables (R : lattice)( A:Type) (y:A) (s:spec A).
Variables (i m : heap) (di :def i) (dm: def m).
Variables (p: s.1 i) (q: s.2 y i m).


Definition st_leq (e1 e2 : SK R s) := forall i, model e1 i <== model e2 i.

(* Order on SK A is defined by lifting the order on the set of results *)
(* for every continuation K *)

Lemma st_refl e : st_leq e e.
Proof. by move=>j; apply: poset_refl. Qed.

Lemma st_trans e1 e2 e3 : st_leq e1 e2 -> st_leq e2 e3 -> st_leq e1 e3.
Proof. by move=>H1 H2 j; apply: poset_trans. Qed.

Lemma st_asym e1 e2 : st_leq e1 e2 -> st_leq e2 e1 -> e1 = e2.
Proof.
move: e1 e2=>[e1][e2]=>E1 E2; congr Prog; apply: fext=>j.
by apply: poset_asym.
Qed.

Definition st_bot : SK R s := Prog (fun i d p _ => bot).

Lemma st_botP e : st_leq st_bot e.
Proof. by move=>x; apply: botP. Qed.

Definition stPosetMixin := PosetMixin st_botP st_refl st_asym st_trans.
Canonical stPoset := Eval hnf in Poset (SK R s) stPosetMixin.
End SKPos.

(* Lattice Structure for SK *)
Section SKLat.
Variables (R : lattice)(A: Type) (y:A) (s:spec A).

Definition st_sup (u : Pred (SK R s)) : SK R s :=
  Prog  (fun i  => sup [Pred t | exists e: SK R s, t = model e i  /\ e \In u]).

Lemma st_supP (u : Pred (SK _ s)) e : e \In u -> st_leq e (st_sup u).
Proof. by move =>/= H h; apply: supP; rewrite InE /=; exists e. Qed.
 
Lemma st_supM (u : Pred (SK _ s)) e : 
        (forall e1, e1 \In u -> st_leq e1 e) -> st_leq (st_sup u) e.
Proof. by move=>H h; apply: supM =>x [f] [->] H1; apply: H. Qed.

Definition stLatticeMixin := LatticeMixin st_supP st_supM. 
Canonical stLattice := Lattice (SK R s) stLatticeMixin.
End SKLat.

Module CCC.

Definition verify A i (s : spec A) (q : A -> heap -> Prop) :=
  def i -> s.1 i /\ forall y m, s.2 y i m -> def m -> q y m.

Definition conseq A (s : spec A) (p : pre) (q : post A) := 
  forall i, p i -> verify i s (fun y m => q y i m).

Local Notation conseq1 := 
  (fun A (s1 s2 : spec A) => 
     conseq s1 (let: (x, _) := s2 in x) 
               (let: (_, x) := s2 in x)).

(*************************)
(* Structural primitives *)
(*************************)

Section Return.
Variables (R : lattice)(A : Type) (x : A).

Definition ret_s : spec A :=
  (fun i => True, fun y i m => y = x /\ m = i).

Lemma retQ i : ret_s.2 x i i.
Proof. by []. Qed.

Definition ret : SK R ret_s :=
  Prog (fun i d pf k=> k x i d (retQ i)).
End Return.
Arguments ret {R A} x.

Section Bind.
Variables (R : lattice)(A B : Type).
Variables (s1 : spec A) (s2 : A -> spec B).
Variables (e1 : SK R s1) (e2 : forall x, SK R (s2 x)).

Definition bind_pre : pre := 
  fun i => 
     s1.1 i /\ forall x m, def m -> s1.2 x i m -> (s2 x).1 m.

Definition bind_post : post B := 
   fun y i m => exists x, exists h, 
     def h /\ s1.2 x i h /\ (s2 x).2 y h m.

Definition bind_s : spec B := (bind_pre, bind_post).

Lemma bindP1 h : bind_s.1 h -> s1.1 h.
Proof. by case. Qed. 

Lemma bindP2 i x h : bind_s.1 i -> def h -> s1.2 x i h -> (s2 x).1 h.
Proof. by move=>[_]; apply. Qed.
 
Lemma bindQ i x h y m : 
        def h -> s1.2 x i h -> (s2 x).2 y h m -> bind_s.2 y i m.
Proof. by move=>*; exists x; exists h. Qed.

Definition bind : SK _ bind_s :=
  Prog (fun i d1 pf1 k => 
          model e1 i d1 (bindP1 pf1) 
           (fun x h d2 p2 => model (e2 x) h d2 (bindP2 pf1 d2 p2) 
           (fun y m d3 p3 => k y m d3 (bindQ d2 p2 p3)))).

End Bind.
Arguments bind {R A B s1 s2} e1 e2.

Lemma conseq_refl A (s : spec A) : conseq1 A s s. 
Proof. by []. Qed.
Hint Resolve conseq_refl.
  
Lemma conseq_trans A (s1 s2 s3 : spec A) : 
  conseq1 A s1 s2 -> conseq1 A s2 s3 -> conseq1 A s1 s3. 
Proof. 
move=>/= H1 H2 i P3i Di; rewrite /conseq in H1 H2.
case: (H2 i P3i Di) => P2i H3; case: (H1 i P2i Di)=> P1i H4; split=>//.
by move=> y m Q1m Dm; move: (H4 y m Q1m Dm)=> /= Q2m; apply: (H3 y m Q2m Dm).
Qed.
Hint Resolve conseq_trans.

Section Consequence.
Variables (R : lattice)(A : Type) (s1 s2 : spec A) (e : SK R s1).
Variables (pf : conseq1 A s1 s2).

Lemma doP i : def i -> s2.1 i -> s1.1 i.
Proof. by case: s2 pf=>p q /= H D P; case: (H _ P D). Qed.

Lemma doQ i : 
  def i -> s2.1 i -> (forall y m, def m -> s2.2 y i m ->  R) -> R.
Proof.
case: s2 pf=>p q /= H D K; case: (H i K D)=>H1 H2 H3.
by apply:(model e i D H1)=>y m D'; move/H2; move/(_ D'); apply: H3 D'.
Qed.

Definition Do : SK R s2 := Prog doQ.
End Consequence.
Arguments Do {R A s1 s2} e pf.
    
(* We define map using bind and ret *)
Section Map.
Variables (R : lattice)(A B: Type) (s: spec A) (f: A -> B) (me: SK R s).
Definition map_s : spec B :=
  ( s.1 ,fun y i m => exists x: A, y = f x /\ s.2 x i m). 

Lemma mapQ : conseq (bind_s s (fun x : A => ret_s (f x))) 
  (let '(x, _) := map_s in x) (let '(_, x) := map_s in x).
Proof.
by  move=> i Pi D; split =>//= y m [x][h][Dh][H1][->]-> _ ; exists x.
Qed.
(* We have to do the proof by hand i.e. w/o the lemmata in srules.v. *)
(* Fortunately, it' s quite silly *)

Program Definition Map : SK _ map_s :=
  @Do R B _ map_s (bind me (fun x => ret (f x))) mapQ.
End Map.
Arguments Map {R A B s} f me.

(* We define join using bind and id *)
Section Join.
Variables (R : lattice) (A: Type).
Variables (s_inner: spec A) (s_outer: spec (SK R s_inner)).

Definition join_s : spec A := bind_s s_outer (fun _ => s_inner ).

Variable me: SK R s_outer. 

Lemma joinQ : conseq (bind_s s_outer (fun _ : SK _ s_inner => s_inner))
  (let '(x, _) := join_s in x) (let '(_, x) := join_s in x).
Proof. by []. Qed.

Program Definition join : SK R join_s :=
  @Do R _ _ join_s (bind me (fun x => x) ) joinQ.
End Join.
Arguments join {R A s_inner s_outer} _.

Section Abort.
Variables (R : lattice)(A B: Type) (sg st : spec A).
Variables (g: Kont R sg) (t: SK R st). 

Definition abort_s : spec B := 
  (fun i=> sg.1 i /\ conseq1 A st (fun j => sg.1 j /\ j = i, sg.2),
    fun y i m => False).

Program
Definition abort : @SK R B abort_s := 
  Prog (fun i d1 p1 k => @g i (Do t _) d1 _).
End Abort.
Arguments abort {R A} B {sg st} g t.

Section Throw.
Variables (R : lattice)(A B : Type) (sg: spec A) (g: Kont R sg) (x : A).

Definition throw_s : spec B := 
  (fun i=> sg.1 i /\ conseq1 A (ret_s x) (fun j => sg.1 j /\ j = i, sg.2), 
    fun y i m => False).

Lemma throwQ :
conseq (abort_s B sg (ret_s x)) (let '(x, _) := throw_s in x)
  (let '(_, x) := throw_s in x).
Proof. by []. Qed.

Program
Definition throw : @SK R B (throw_s)  := 
 @Do R B _ throw_s (@abort R A B sg (ret_s x) g (ret x)) throwQ.
End Throw.
Arguments throw {R A B sg} g x.

(* CallCC implementations that arise from the algebraic constraints:          *)

(* qcc qf qe ; q2 = qcc (qf ; q2) (q1 ; q2)                                   *)
 
(* \ i . pcc i /\ (forall y m . qcc y i m -> p2 m) =                          *)
(* pcc (\i. pf i /\ forall y m . qf y i m -> p2 m)                            *)
(*     (\i. pe i /\  forall y m . qe y i m -> p2 m)                           *)  

(* The first solution is pcc = pf /\ pe qcc = qf \/ qe *)
(* However this implies that callcc is a rollback                             *)

(* The second solution involves keeping track of changes in R.                *)
(* R : heap -> heap -> prop                                                   *)


(* Relational Operators *)
Definition alg_spec (A: Type )(s1 s2: spec A) : spec A:=
  (fun i => s1.1 i /\ s2.1 i , fun y i m => s1.2 y i m \/ s2.2 y i m).

Section RelOps.
Variables (A: Type)  (r: heap -> heap -> Prop) (pf: pre) (qf: post A).

Definition precomp : pre :=
  fun i => forall h, def  h ->  r i h -> pf h.

Definition postcomp : post A :=
  fun y i m => exists h, def h /\ r i h /\ qf y h m. 
End RelOps.

Section CallCCR.
Variables (X: lattice)( A : Type).
Variables (R: heap -> heap -> Prop) (sf se:  heap -> spec A).
Variable (h: forall j,  
  Kont X (fun i => R j i /\ (sf j).1 i, (sf j).2) -> SK X (se j)).

Definition cccR_s : spec A := 
   (fun i =>  (se i).1 i /\ forall s, R i s -> def s -> (sf i).1 s, 
    fun y i m => (se i).2 y i m \/ exists s, R i s /\ def s /\ (sf i).2 y s m).

Section Conts.

Variables (i : heap) (s: heap) (D : def i) (P1 : cccR_s.1 i).
Variable  (Ps: R i s /\ (sf i).1 s).
Variable (K : forall y m, def m -> cccR_s.2 y i m ->  X).

Lemma cccRPe : (se i).1 i. 
Proof. by case: P1. Qed.

Lemma cccRQf (Ds: def s) y m :def m -> (sf i).2 y s m -> X.
Proof.
move=> Dm H0; apply: (@K y m Dm); right. 
by case: P1 Ps => /= H1 H2 [H3 _]; exists s ; repeat split=>//.
Qed.

Lemma cccRQe y m : def m -> (se i).2 y i m ->  X.
Proof. by  move => Dm H1; apply: (@K y m Dm); left. Qed.
End Conts.

Program
Definition callccR : SK X (cccR_s) :=
  Prog (fun i d1 p1 k => 
     model (@h i (fun s m d2 ps => model m s d2 _ (cccRQf p1 ps k d2)  ))
                                                i d1 (cccRPe p1) (cccRQe k) ).

End CallCCR.
Arguments callccR {X A} R sf {se} h.

(* Explicit P_f = True *)
Section CallCC.
Variables (X: lattice) (A : Type) (R: heap -> heap -> Prop).
Variables (Qf : heap -> post A) (se:  heap -> spec A). 

Definition ccc_T := 
  CCC.cccR_s  R (fun i => (xPredT, Qf i)) se.

Program
Definition callcc := 
  @callccR X A  R (fun i => (xPredT, Qf i)) .
End CallCC.
Arguments callcc {X A} R Qf {se} h.

Section Fix. 
Variables (R : lattice)(A : Type) (B : A -> Type) (s : forall x, spec (B x)).
Notation tp := (forall x, SK R (s x)).
Notation lat := (dfunLattice (fun x => [lattice of SK R (s x)])).
Variable (f : tp -> tp).

(* we take a fixpoint not of f, but of its monotone completion f' *)
(* should eventually prove that f' is monotone *)

Definition f' (e : lat) := 
  sup [Pred t : lat | exists e', e' <== e /\ t = f e'].

Definition ffix : tp := tarski_lfp f'. 
End Fix.
Arguments ffix {R A B s} f x.

(***********************)
(* the non-structurals *)
(***********************)

Section Allocation.
Variables (R : lattice)(A : Type) (x : A).

Definition alloc_s : spec ptr := 
  (fun i => True, fun y i m => m = y :-> x :+ i).

Lemma allocD i : def i -> def (fresh i :-> x :+ i).
Proof. by move=>D; rewrite defPtUn D fresh_null -[fresh i]ptr0 dom_fresh. Qed.

Lemma allocQ i y : alloc_s.2 y i (y :-> x :+ i).
Proof. by []. Qed.

Definition alloc : SK R alloc_s :=
 Prog (fun i d pf k => 
        let: y := fresh i in
          k y (y :-> x :+ i) (allocD d) (allocQ i y)).

End Allocation.

Arguments alloc {R A} x.

Section BlockAllocation.
Variables (R : lattice)(A : Type) (v : A) (n : nat).

Definition allocb_s : spec ptr :=  
   (fun i => True, 
    fun y i m => exists r, y = r /\ m = i :+ updi r (nseq n v)).


Lemma allocbD i : def i -> def (i :+ updi (fresh i) (nseq n v)).
Proof.
move=>D; elim:n=>[|k IH]; first by rewrite unh0 /=.
rewrite (_ : nseq k.+1 v = rcons (nseq k v) v); last first.
- by elim: {IH} k=>[|k IH] //=; rewrite -IH.
rewrite updi_last unA unC defPtUn IH /=. 
rewrite ptr_null negb_and fresh_null /=.
rewrite domUn !inE /= negb_and IH negb_or /=.
by rewrite dom_fresh updimV negb_and fresh_null ltnn.
Qed.

Lemma allocbQ i r : allocb_s.2 r i (i :+ updi r (nseq n v)).
Proof. by exists r. Qed.

Program
Definition allocb : SK R allocb_s :=
 Prog (fun i d pf k =>
        let: r := fresh i in
          k r (i :+ updi r (nseq n v)) (allocbD d) (allocbQ i r)).

End BlockAllocation.
Arguments allocb {R A} v n.

Section Deallocation.
Variables  (R : lattice) (x : ptr).

Definition dealloc_s : spec unit := 
  (fun i => exists A, exists v : A, exists h, i = x :-> v :+ h, 
   fun y i m => m = free x i).

Lemma deallocD i : def i -> def (free x i).
Proof. by move=>D; rewrite defF. Qed.

Lemma deallocQ i : dealloc_s.2 tt i (free x i).
Proof. by []. Qed.

Definition dealloc : SK R dealloc_s :=
  Prog (fun i d pf k =>
          k tt (free x i) (deallocD d) (deallocQ i)).
End Deallocation.
Arguments dealloc {R} x.

Section Read.
Variables (R : lattice)(A : Type) (x : ptr).

Definition read_s : spec A := 
  (fun i => exists v : A, exists h, i = x :-> v :+ h, 
   fun y i m => m = i /\ forall v h, i = x :-> v :+ h -> y = v). 

Lemma xtp i : def i -> read_s.1 i -> 
Dyn.typ (look x i) = A.
Proof. by move=>D [v][h] E; rewrite {}E in D *; rewrite lookPtUn. Qed.

Definition xval i (d : def i) (pf : read_s.1 i) : A :=
  coerce (T:=id) (Dyn.val (look x i)) (xtp d pf). 

Lemma readQ i (d : def i) (pf : read_s.1 i) : read_s.2 (xval d pf) i i.
Proof.
split=>//; move=>v h E; rewrite {}E in d pf *.
rewrite /xval; move: (xtp d pf)=>F.
by move: (F); rewrite lookPtUn //; move=>{F} F; rewrite eqc.
Qed.

Definition read : SK R read_s :=
  Prog (fun i d pf k => k (xval d pf) i d (readQ d pf)). 

End Read.
Arguments read {R A} x.

Section Write.
Variables (R : lattice)(A : Type) (x : ptr) (v : A).

Definition write_s : spec unit := 
  (fun i => exists B, exists w : B, exists h, i = x :-> w :+ h, 
   fun y i m => forall B (w : B) h, 
      i = x :-> w :+ h -> m = x :-> v :+ h /\ y = tt ).

Definition writeD i : def i -> write_s.1 i -> def (upd i x (dyn v)).
Proof. 
move=>d [B][w][h] E; rewrite {}E in d *.
by rewrite defU d (defPt_null d).
Qed.

Definition writeQ i : write_s.2 tt i (upd i x (dyn v)). 
Proof. by move=>B w h ->; rewrite updPtUn. Qed.

Definition write : SK R write_s :=
  Prog (fun i d pf k => k tt (upd i x (dyn v)) (writeD d pf) (@writeQ i)).
End Write.
Arguments write {R A} x v.

 (****************)
(* Conditionals *)
(****************)

(* We need at least, conditionals for boolean or nat *)

Section If.
Variables (R : lattice)(A : Type) (b : bool) (s1 s2 : spec A). 

Program
Definition If (e1 : SK R s1) (e2 : SK R s2) :
              SK R (if b then s1 else s2) :=
          match b with true => Do e1 _ | false => Do e2 _ end.
End If.

Section CondNat.
Variables (R : lattice) (A : Type) (n : nat) (s1 : spec A) (s2 : nat -> spec A).

Program 
Definition Match_nat (e1 : SK R s1) (e2 : forall n, SK R (s2 n)) :
             SK _ (match n with 0 => s1 | m.+1 => s2 m end) :=
  match n with 0 => Do e1 _ | m.+1 => Do (e2 m) _ end.
End CondNat.

Section CondSeq.
Variables (R : lattice)(A B: Type) (xs : seq B).
Variables (s1 : spec A) (s2 : B -> seq B-> spec A).

Program
Definition Match_seq (e1 : SK R s1) (e2 : forall x xs, SK R (s2 x xs)) :
             SK R (match xs with [::] => s1 | y :: ys => s2 y ys end) :=
  match xs with [::] => Do e1 _ | y :: ys => Do (e2 y ys) _ end.
End CondSeq.

Section  Option.
Variables (R : lattice) (A B: Type).
Variables (o : option A) (s2: spec B) (s1: A -> spec B).

Program
Definition Match_opt (e1 : forall p, SK R (s1 p)) (e2 : SK R s2) :
  SK R (if o is (Some p) then  (s1 p) else s2 ) :=
  match o with Some p => Do (e1 p) _ | None => Do e2 _ end.
End Option.

Arguments If {R A} b {s1 s2} e1 e2.
Arguments Match_nat {R A} n {s1 s2} e1 e2.
Arguments Match_seq {R A B} xs {s1 s2} e1 e2.
Arguments Match_opt {R A B} o {s2 s1} e1 e2.

Section Let.
Variable (R : lattice)(A B: Type) (s2: A -> spec B).

Program
Definition MLet (exp : A) (e2 : forall p: A, SK R (s2 p)) 
  : SK R (let: p:= exp in  (s2 p)) 
    :=
    Do ( bind (ret exp) e2) _.
Next Obligation.
move=>/= i P; split=>/=; first by split=>// x m Dm [-> ->].
by move=>y m [x][h][Dh][[-> ->]].
Qed.
End Let.
Arguments MLet {R A B s2} exp e2.
End CCC.

(* HTT notation *)
Notation "'Do' e" := (CCC.Do e _) (at level 80) : ccscope.
Notation "'callccAlg'" := (CCC.callccR) (at level 80) : ccscope.
Notation "'callcc'" := (CCC.callcc) (at level 80) : ccscope.
Notation "x '<--' c1 ';' c2" := (CCC.bind c1 (fun x => c2))
  (at level 78, right associativity) : ccscope.
Notation "c1 ';;' c2" := (CCC.bind c1 (fun _ => c2))
  (at level 78, right associativity) : ccscope.
Notation "'!' x" := (CCC.read x) (at level 50) : ccscope.
Notation "e1 '::=' e2" := (CCC.write e1 e2) (at level 60) : ccscope.
Notation "'MLet' exp 'In' rest" := (CCC.MLet exp rest) (at level 75) : ccsope.
Notation "'If' b 'Then' e1 'Else' e2" :=
  (CCC.If b e1 e2) (at level 80) : ccscope.
Notation "'Match_nat' n e1 e2" :=
  (CCC.Match_nat n e1 e2) (at level 80) : ccscope.
Notation "'Match_seq' xs e1 e2" :=
  (CCC.Match_seq xs e1 e2) (at level 80) : ccscope.
Notation "'Match_opt' p e1 e2" :=
  (CCC.Match_opt p e1 e2) (at level 80) : ccscope.
Notation "'alloc' n" := (CCC.alloc n) (at level 80) : ccscope.
Notation "'dealloc'" := (CCC.dealloc) (at level 80) : ccscope.
Notation "'allocb' n m" := (CCC.allocb n m) (at level 80) : ccscope.
Notation "'ret' x" := (CCC.ret x) (at level 80) : ccscope.
Notation "'throw'":= (CCC.throw) (at level 80) : ccscope.
Notation "'ffix'" := (CCC.ffix) (at level 80) : ccscope.
Notation "'abort'":= (CCC.abort) (at level 80) : ccscope.
Notation "'join'":= (CCC.join) (at level 80) : ccscope.
