(************************************)
(* ST recursive model for callcc    *)
(*             WHOLE HEAP           *)
(************************************)

From mathcomp Require Import ssreflect ssrfun ssrbool ssrnat eqtype seq.
From Kont Require Import rels domains prelude heaps hprop.
Require Import Eqdep.

Set Implicit Arguments.
Unset Strict Implicit.
Unset Printing Implicit Defensive.

Delimit Scope ccscope with cc. 
Open Scope ccscope.

(*****************)
(* Program model *)
(*****************)

Definition pre := heap -> Prop.
Definition post A := A -> heap -> heap -> Prop.
Definition spec A := (pre * post A)%type.

Definition prog A (s : spec A) := 
  forall i, def i -> s.1 i -> 
    (forall y m, def m -> s.2 y i m -> Pred False) -> Pred False.

Structure ST A (s : spec A) : Type := Prog { model : prog s}.

Arguments ST {A}.
Arguments model {A s}.

(* The type of continuations of type A and preconditon p *)
Definition cont_s A p (y : A) : spec False := 
  (p y, fun y i m => False).

Definition STcont A (p : A -> heap -> Prop) : Type :=
  forall y, ST (cont_s p y).

(* In order  to be able to flatten pre and post conditions when returning     *)
(* computations will attempt to define it as a particular return type which   *)
(* carries its pre and post conditions, by means of a dependent sum type:     *)

Inductive SigST (A : Type) := BigST {spec_of : spec A; comm: ST spec_of}.


(* Two computations are equal if they have the same model:          *)
(* The lemma is just extentional equality extended to all arguments *)

Lemma modelE A (s: spec A) (e1 e2: ST s) :
        e1 = e2 <-> model e1 = model e2.
Proof. by move: e1 e2=>[e1][e2]; split=>/= [[]|->]. Qed.

(**************************)
(* Poset Structure for ST *)
(**************************)
Section STPos.
Variables (A:Type) (y:A) (s:spec A).
Variables (i m : heap) (di :def i) (dm: def m).
Variables (p: s.1 i) (q: s.2 y i m).
Variable ( k: forall (x : A) (h : heap),
               def h -> s.2 x i h -> Pred False).


Definition st_leq (e1 e2 : ST s) := forall i, model e1 i <== model e2 i.

(* Order on ST A is defined by lifting the order on the set of results *)
(* for every continuation K *)

Lemma st_refl e : st_leq e e.
Proof. by move=>j; apply: poset_refl. Qed.

Lemma st_trans e1 e2 e3 : st_leq e1 e2 -> st_leq e2 e3 -> st_leq e1 e3.
Proof. by move=>H1 H2 j; apply: poset_trans. Qed.


Lemma st_asym e1 e2 : st_leq e1 e2 -> st_leq e2 e1 -> e1 = e2.
Proof.
move: e1 e2=>[e1][e2]=>E1 E2.
congr Prog; apply: fext=>j.
by apply: poset_asym.
Qed.

Definition st_bot : ST s := Prog (fun i d p _ => xPred0).

Lemma st_botP e : st_leq st_bot e.
Proof. by move=>x; apply: botP. Qed.

Definition stPosetMixin := PosetMixin st_botP st_refl st_asym st_trans.
Canonical stPoset := Eval hnf in Poset (ST s) stPosetMixin.
End STPos.


(* Lattice Structure for ST *)
Section STLat.
Variables (A: Type) (y:A) (s:spec A).


(* Definition st_sup (u : Pred (@ST _ s)) : @ST A s :=  *)
(*   @Prog A s (fun i =>  *)
(*                sup (@SimplPred (def i -> *)
(*                           s.1 i -> *)
(*                           (forall (y : A) (m : heap), *)
(*                              def m -> s.2 y i m -> Pred False) -> *)
(*                           Pred False) *)
(*                    (fun t => exists e : ST s , t = model e i /\ e \In u))). *)

Definition st_sup (u : Pred (ST s)) : ST s :=
  Prog (fun i => sup [ Pred t | exists e : ST s , t = model e i /\ e \In u]).



Lemma st_supP (u : Pred (ST s)) e : e \In u -> st_leq e (st_sup u).
Proof. by move => H h; apply: supP; rewrite InE /=; exists e. Qed.
 
Lemma st_supM (u : Pred (ST s)) e : 
        (forall e1, e1 \In u -> st_leq e1 e) -> st_leq (st_sup u) e.
Proof. by move=>H h; apply: supM =>x [f] [->] H1; apply: H. Qed.

Definition stLatticeMixin := LatticeMixin st_supP st_supM. 
Canonical stLattice := Lattice (ST s) stLatticeMixin.

End STLat.


Module CCC.


Definition verify A i (s : spec A) (q : A -> heap -> Prop) :=
  def i -> s.1 i /\ forall y m, s.2 y i m -> def m -> q y m.

Definition conseq A (s : spec A) (p : pre) (q : post A) := 
  forall i, p i -> verify i s (fun y m => q y i m).

Local Notation conseq1 := 
  (fun A (s1 s2 : spec A) => 
     conseq s1 (let: (x, _) := s2 in x) 
               (let: (_, x) := s2 in x)).


(*************************)
(* Structural primitives *)
(*************************)

Section Return.
Variables (A : Type) (x : A).

Definition ret_s : spec A :=
  (fun i => True, fun y i m => y = x /\ m = i).

Lemma retQ i : ret_s.2 x i i.
Proof. by []. Qed.

Definition ret : ST ret_s :=
  Prog (fun i d pf k=> k x i d (retQ i)).
End Return.
Arguments ret {A}.

Section Throw.
Variables (A : Type) (p : A -> heap -> Prop) (f : STcont p) (x : A).

Definition throw_s B : spec B := (p x, fun y i m => False).

Program
Definition throw B : ST (throw_s B) :=
  Prog (fun i d1 (p1:(throw_s B).1 i) k => model (f x) i d1 p1 _ ). 
Next Obligation. by []. Qed.
End Throw.
Arguments throw {A p} f x.

Section Bind.
Variables (A B : Type).
Variables (s1 : spec A) (s2 : A -> spec B).
Variables (e1 : ST s1) (e2 : forall x, ST (s2 x)).

Definition bind_pre : pre := 
  fun i => 
     s1.1 i /\ forall x m, def m -> s1.2 x i m -> (s2 x).1 m.

Definition bind_post : post B := 
   fun y i m => exists x, exists h, 
     def h /\ s1.2 x i h /\ (s2 x).2 y h m.


Definition bind_s : spec B := (bind_pre, bind_post).

Lemma bindP1 h : bind_s.1 h -> s1.1 h.
Proof. by case. Qed. 

Lemma bindP2 i x h : bind_s.1 i -> def h -> s1.2 x i h -> (s2 x).1 h.
Proof. by move=>[_]; apply. Qed.
 
Lemma bindQ i x h y m : 
        def h -> s1.2 x i h -> (s2 x).2 y h m -> bind_s.2 y i m.
Proof. by move=>*; exists x; exists h. Qed.

Definition bind : ST bind_s :=
  Prog (fun i d1 pf1 k => 
          model e1 i d1 (bindP1 pf1) 
           (fun x h d2 p2 => model (e2 x) h d2 (bindP2 pf1 d2 p2) 
           (fun y m d3 p3 => k y m d3 (bindQ d2 p2 p3)))).

End Bind.
Arguments bind [A B s1 s2].

Lemma conseq_refl A (s : spec A) : conseq1 A s s. 
Proof. by []. Qed.

Hint Resolve conseq_refl.
Lemma conseq_trans A (s1 s2 s3 : spec A) : 
  conseq1 A s1 s2 -> conseq1 A s2 s3 -> conseq1 A s1 s3. 
Proof. 
move=>/= H1 H2 i P3i Di; rewrite /conseq in H1 H2.
case: (H2 i P3i Di) => P2i H3; case: (H1 i P2i Di)=> P1i H4; split=>//.
by move=> y m Q1m Dm; move: (H4 y m Q1m Dm)=> /= Q2m; apply: (H3 y m Q2m Dm).
Qed.
Hint Resolve conseq_trans.

Section Consequence.
Variables (A : Type) (s1 s2 : spec A) (e : ST s1).
Variables (pf : conseq1 A s1 s2).

Lemma doP i : def i -> s2.1 i -> s1.1 i.
Proof. by case: s2 pf=>p q /= H D P; case: (H _ P D). Qed.

Lemma doQ i : 
       def i -> s2.1 i -> (forall y m, def m -> s2.2 y i m -> Pred False) -> 
       Pred False.
Proof.
case: s2 pf=>p q /= H D K; case: (H i K D)=>H1 H2 H3.
by apply: (model e i D H1)=>y m D'; move/H2; move/(_ D'); apply: H3 D'.
Qed.

Definition Do : ST s2 := Prog doQ.

End Consequence.

Section CallCC.
Variables (A : Type) (s : spec A) (s' : heap -> spec A). 
Variables (f : forall j, STcont (s.2^~ j) -> ST (s' j)).
Variables (pf : forall j, conseq1 A (s' j) (fun i => j = i /\ s.1 i, s.2)).

Definition callcc_s : spec A := s.

Section Conts.
Variables (i : heap) (D : def i) (H : callcc_s.1 i).
Variables (K : forall y m, def m -> callcc_s.2 y i m -> Pred False).

Lemma cc2P1e j : def j -> callcc_s.1 j -> (s' j).1 j.
Proof. by move=>Dj Pj; case: (@pf j j _ ). Qed.

Lemma cc2Q1e y m : def m -> (s' i).2 y i m -> Pred False.
Proof.
move=>D'; case: (@pf i i (conj (erefl i) H) D)=>_ H1.
by move=>H2; apply: (@K y m D'); apply: H1 H2 D'.
Qed.

Lemma cc2Q2e y m : def m -> (cont_s (s.2^~ i) y).1 m -> Pred False.
Proof. move=>Dm; apply: (@K y m Dm). Qed.

End Conts.

Program
Definition callcc : ST (callcc_s) :=
  Prog (fun i d1 p1 k => 
         model (f (fun y => Prog (fun h d2 p2 _ => cc2Q2e k d2 p2)))
         i d1 (cc2P1e d1 p1) (cc2Q1e d1 p1 k )).


End CallCC.
Arguments callcc [A] s [s'] _ _.

Section Fix. 
Variables (A : Type) (B : A -> Type) (s : forall x, spec (B x)).
Notation tp := (forall x, ST (s x)).
Notation lat := (dfunLattice (fun x => [lattice of ST (s x)])).
Variable (f : tp -> tp).

(* we take a fixpoint not of f, but of its monotone completion f' *)
(* should eventually prove that f' is monotone *)

Definition f' (e : lat) := 
  sup [Pred t : lat | exists e', e' <== e /\ t = f e'].

Definition ffix : tp := tarski_lfp f'. 
End Fix.


(***********************)
(* the non-structurals *)
(***********************)

Section Allocation.
Variables (A : Type) (x : A).

Definition alloc_s : spec ptr := 
  (fun i => True, fun y i m => m = y :-> x :+ i).

Lemma allocD i : def i -> def (fresh i :-> x :+ i).
Proof. by move=>D; rewrite defPtUn D fresh_null -[fresh i]ptr0 dom_fresh. Qed.

Lemma allocQ i y : alloc_s.2 y i (y :-> x :+ i).
Proof. by []. Qed.

Definition alloc : ST alloc_s :=
 Prog (fun i d pf k => 
        let: y := fresh i in
          k y (y :-> x :+ i) (allocD d) (allocQ i y)).

End Allocation.
Arguments alloc [A].

Section BlockAllocation.
Variables (A : Type) (v : A) (n : nat).

Definition allocb_s : spec ptr :=  
   (fun i => True, 
    fun y i m => exists r, y = r /\ m = i :+ updi r (nseq n v)).


Lemma allocbD i : def i -> def (i :+ updi (fresh i) (nseq n v)).
Proof.
move=>D; elim: n => [|k IH ]; first by rewrite unh0 /=.
rewrite (_ : nseq k.+1 v = rcons (nseq k v) v); last first.
- by elim: {IH} k=>[|k IH] //=; rewrite -IH.
rewrite updi_last unA unC defPtUn IH /=. 
rewrite ptr_null negb_and fresh_null /=.
rewrite domUn !inE /= negb_and IH negb_or /=.
by rewrite dom_fresh updimV negb_and fresh_null ltnn.
Qed.

Lemma allocbQ i r : allocb_s.2 r i (i :+ updi r (nseq n v)).
Proof. by exists r. Qed.

Program
Definition allocb : ST allocb_s :=
 Prog (fun i d pf k =>
        let: r := fresh i in
          k r (i :+ updi r (nseq n v)) (allocbD d) (allocbQ i r)).

End BlockAllocation.


Section Deallocation.
Variable x : ptr.

Definition dealloc_s : spec unit := 
  (fun i => exists A, exists v : A, exists h, i = x :-> v :+ h, 
   fun y i m => m = free x i).

Lemma deallocD i : def i -> def (free x i).
Proof. by move=>D; rewrite defF. Qed.

Lemma deallocQ i : dealloc_s.2 tt i (free x i).
Proof. by []. Qed.

Definition dealloc : ST dealloc_s :=
  Prog (fun i d pf k =>
          k tt (free x i) (deallocD d) (deallocQ i)).

End Deallocation.
Arguments dealloc x: clear implicits.

Section Read.
Variables (A : Type) (x : ptr).

Definition read_s : spec A := 
  (fun i => exists v : A, exists h, i = x :-> v :+ h, 
   fun y i m => m = i /\ forall v h, i = x :-> v :+ h -> y = v). 

Lemma xtp i : def i -> read_s.1 i -> 
Dyn.typ (look x i) = A.
Proof. by move=>D [v][h] E; rewrite {}E in D *; rewrite lookPtUn. Qed.

Definition xval i (d : def i) (pf : read_s.1 i) : A :=
  coerce (T:=id) (Dyn.val (look x i)) (xtp d pf). 

Lemma readQ i (d : def i) (pf : read_s.1 i) : read_s.2 (xval d pf) i i.
Proof.
split=>//; move=>v h E; rewrite {}E in d pf *.
rewrite /xval; move: (xtp d pf)=>F.
by move: (F); rewrite lookPtUn //; move=>{F} F; rewrite eqc.
Qed.

Definition read : ST read_s :=
  Prog (fun i d pf k => k (xval d pf) i d (readQ d pf)). 

End Read.
Arguments read [A].

Section Write.
Variables (A : Type) (x : ptr) (v : A).

Definition write_s : spec unit := 
  (fun i => exists B, exists w : B, exists h, i = x :-> w :+ h, 
   fun y i m => forall B (w : B) h, 
      i = x :-> w :+ h -> m = x :-> v :+ h /\ y = tt ).

Definition writeD i : def i -> write_s.1 i -> def (upd i x (dyn v)).
Proof. 
move=>d [B][w][h] E; rewrite {}E in d *.
by rewrite defU d (defPt_null d).
Qed.

Definition writeQ i : write_s.2 tt i (upd i x (dyn v)). 
Proof. by move=>B w h ->; rewrite updPtUn. Qed.

Definition write : ST write_s :=
  Prog (fun i d pf k => k tt (upd i x (dyn v)) (writeD d pf) (@writeQ i)).

End Write.
Arguments write [A].

 (****************)
(* Conditionals *)
(****************)

(* We need at least, conditionals for boolean or nat *)

Section If.
Variables (A : Type) (b : bool) (s1 s2 : spec A). 

Program Definition If (e1 : ST s1) (e2 : ST s2) : ST (if b then s1 else s2) :=
  match b with true => Do e1 _ | false => Do e2 _ end.
End If.

Section CondNat.
Variable (A : Type) (n : nat) (s1 : spec A) (s2 : nat -> spec A).

Program 
Definition Match_nat (e1 : ST s1) (e2 : forall n, ST (s2 n)) :
             ST (match n with 0 => s1 | m.+1 => s2 m end) :=
  match n with 0 => Do e1 _ | m.+1 => Do (e2 m) _ end.
End CondNat.

End CCC.

(* HTT notation *)
Notation "'Do' e" := (CCC.Do e _) (at level 80) : ccscope.
Notation "'callcc'" := (CCC.callcc) (at level 80) : ccscope.
Notation "x '<--' c1 ';' c2" := (CCC.bind c1 (fun x => c2))
  (at level 78, right associativity) : ccscope.
Notation "c1 ';;' c2" := (CCC.bind c1 (fun _ => c2))
  (at level 78, right associativity) : ccscope.
Notation "'!' x" := (CCC.read x) (at level 50) : ccscope.
Notation "e1 '::=' e2" := (CCC.write e1 e2) (at level 60) : ccscope.
Notation "'If' b 'Then' e1 'Else' e2" :=
  (CCC.If b e1 e2) (at level 80) : ccscope.
(* Notation "'Case' n '[' e1 '||' e2 ']'" := *)
(*   (CCC.Match_nat n e1 e2) (at level 80) : ccscope. *)
Notation "'alloc' n" := (CCC.alloc n) (at level 80) : ccscope.
Notation "'dealloc'" := (CCC.dealloc) (at level 80) : ccscope.
Notation "'allocb' n m" := (CCC.allocb n m) (at level 80) : ccscope.
Notation "'ret' x" := (CCC.ret x) (at level 80) : ccscope.
Notation "'throw'" := (CCC.throw) (at level 80) : ccscope.
Notation "'ffix'" := (CCC.ffix) (at level 80) : ccscope.



